from twisted.internet.protocol import ServerFactory, Factory, Protocol
from twisted.protocols import basic
from twisted.internet import reactor, defer
from simplejson import loads
from pyf.transport.packets import encode_packet_flow
import Queue

from tgscheduler import scheduler
from pyf.transport import decode_packet_flow, Packet

class BadFlowEnding(Exception):
    pass

class ServerError(Exception):
    pass

class PacketGeneratorAdapter(object):
    def __init__(self, generator):
        self.generator = iter(generator)
        self.inited = False
    
    def __iter__(self):
        return self
    
    def next(self):
        if self.inited:
            return self.adapted_generator.next()
        
        else:
            self.header_str = self.generator.next()
            self.header = Packet(data=loads(self.header_str),
                                 data_type="serialized")
            
            self.adapted_generator = decode_packet_flow(self.generator,
                                        decompress=self.header.get('compressed',
                                                                   False),
                                        pure_flow=True)
            self.inited = True
            return self.header        

class GeneratorAdapter(object):
    def __init__(self, client):
        self.client = client
        
    def __iter__(self):
        queue = self.client.data_queue
        i = 0
        while True:
            if self.client.data_queue.empty() and self.client.data_queue.finished:
                if self.client.data_queue.complete:
                    raise StopIteration
                else:
                    raise BadFlowEnding, \
                          'Bad ending in flow. Corrupted connection ?'
            item = queue.get()
            yield item
            i += 1
    
class FlowReceiver(basic.LineReceiver):
    delimiter='\r\n\0'
    
class DataFlowQueue(Queue.Queue):
    def __init__(self, *args, **kwargs):
        Queue.Queue.__init__(self, *args, **kwargs)
        self.finished = False
        self.complete = False

class FlowProtocol(FlowReceiver):
    go = Packet(dict(message='go', type='control'))
    nogo = Packet(dict(message='nogo',
                       type='control',
                       detail='error in initialization'))
    good = "good."
    ok = Packet(dict(message='ok', type='info'))
    
    def lineReceived(self, line):
        
        if line == self.good:
            self.data_queue.finished = True
            self.data_queue.complete = True
        else:
            self.data_queue.put(line)
#            self.message(self.ok)

    def message(self, message):
        if isinstance(message, Packet):
            for v in encode_packet_flow([message], separator=self.delimiter):
                self.transport.write(v)
        else:
            self.transport.write(message + self.delimiter)
        
    def connectionMade(self):
        try:
            self.factory.numProtocols = self.factory.numProtocols+1
            self.factory.clients.append(self)
            self.data_queue = DataFlowQueue(maxsize=self.factory.max_items)
            
            if self.factory.numProtocols > self.factory.max_clients:
                self.error(ServerError('Too many connections. Try again later.'))
            else:   
                flow = PacketGeneratorAdapter(GeneratorAdapter(self))
                scheduler.add_single_task(self.factory.data_handler,
                                          kw=dict(flow=flow,
                                                  client=self),
                                          initialdelay=0)
                self.message(self.go)
        except Exception, e:
            self.transport.loseConnection()
            raise e

    def connectionLost(self, reason):
        self.factory.clients.remove(self)
        self.factory.numProtocols = self.factory.numProtocols-1
        self.data_queue.finished = True
        
    def error(self, error):
        self.message(Packet(dict(type='error',
                                 message=str(error),
                                 error_type=type(error).__name__)))
        self.transport.loseConnection()
    
    def success(self, message):
        self.message(Packet(dict(type='success', message=message)))
        self.transport.loseConnection()
        
        
class FlowServer(ServerFactory):
    protocol = FlowProtocol
    
    def __init__(self, data_handler, max_items=1000, max_clients=100):
        self.max_items = max_items
        self.max_clients = max_clients
        self.data_handler = data_handler
        self.numProtocols = 0
        self.clients = list()

def launch_server():
    """This runs the protocol on port 8000"""
    scheduler.start_scheduler()
    
    def toto_plouf(flow, client=None):
        for i, item in enumerate(flow):
            if not i%50:
                print item.num
            
        print "end of flow..."
    
    factory = FlowServer(toto_plouf)
    reactor.listenTCP(8000,factory)
    scheduler.add_single_task(reactor.run,
                              kw=dict(installSignalHandlers=0),
                              initialdelay=0)
    while True:
        pass

# this only runs if the module was *not* imported
if __name__ == '__main__':
    launch_server()


from pyf.transport import encode_packet_flow, Packet
from pyf.transport.packets import decode_packet_flow
from pyf.splitter.inputsplitter import tokenize
import sys
import time
import socket
from itertools import chain
from pyjon.events import EventDispatcher

class ServerError(Exception):
    pass

class StationClient(object):
    __metaclass__ = EventDispatcher
    
    finished_message = 'good.'
    
    def __init__(self, host, port=6789,
                 waits_for_success=False,
                 separator='\r\n\0'):
        self.host = host
        self.port = port
        self.waits_for_success = waits_for_success
        self.separator = separator
        
    def __get_socket(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.host, self.port))
        
        return s
    
    def handle_message(self, sock, item):
        self.emit_event('message_received',
                        item)
        if item.type == 'error':
            sock.close()
            self.emit_event('failure', item)
            raise ServerError, item.message
        
        elif item.type == 'appinfo':
            self.emit_event('appinfo',
                            item.get('key'),
                            item.get('value'))
            
        elif item.type == 'success':
            self.emit_event('success', item)
            
        else:
            self.emit_event('other',
                            item)
        
    def receive_data(self, sock, remaining=""):
        sep = self.separator
        
        try:
            out = remaining + sock.recv(1024)
        except socket.error:
            out = remaining
            
        if out:
            if sep not in out:
                remaining += out
            else:
                items, remaining = tokenize(out, sep)
                
                return remaining, decode_packet_flow(iter(items),
                                                     separator=sep,
                                                     pure_flow=True)
        
        else:
            return remaining, []

    def call(self, content, header=None):
        s = self.__get_socket()
        
        if header is not None:
            if isinstance(header, dict):
                header = Packet(header)
                
            content = chain([header], content)
        
        remaining = ""
        sep = self.separator
        for i, data in enumerate(encode_packet_flow(content,
                                       separator=sep,
                                       compress=False)):
            if not i % 2:
                yield True
                
            s.sendall(data)
            
            s.setblocking(False)
            
            remaining, out_items = self.receive_data(s, remaining)
            for item in out_items:
                self.handle_message(s, item)
                yield item
                
#            try:
#                out = remaining + s.recv(1024)
#            except socket.error:
#                out = remaining
#                
#            if out:
#                if sep not in out:
#                    remaining += out
#                
#                else:
#                    items, remaining = tokenize(out, sep)
#                    
#                    for item in decode_packet_flow(iter(items),
#                                                   separator=sep,
#                                                   pure_flow=True):
#                        self.handle_message(sock, item)
#                        
#                        yield item
            s.setblocking(True)
        
        if self.waits_for_success:
            finished = False
            fsent = False
            s.setblocking(False)
            
            while not finished:
                if not fsent:
                    try:
                        s.sendall(self.finished_message + sep)
                        fsent = True
                    except socket.error:
                        pass
                
                remaining, out_items = self.receive_data(s, remaining)
                for item in out_items:
                    self.handle_message(s, item)
                    if item.get('type') == 'success':
                        finished = True
                        
                    yield item
        else:
            s.sendall(self.finished_message + sep)
            
        s.close()

if __name__ == '__main__':
    client = StationClient('127.0.0.1', 6789, True)
    
    def message_handler(message_packet):
        print message_packet
    
    client.add_listener('message_received', message_handler)
    flow = (Packet(dict(Field1=i+1,
                        Field2=('titi', 'tata')[i%2], num=i+1,
                        Field3=(i+1)*10))
            for i in range(10000))
    
    values = client.call(
         flow,
         header=dict(authtkt='125fa7e6398fdede0cdbfb700251ae934bffbaeemanager!',
                     action='launch_tube',
                     object_name='simple'))
    
    for i, value in enumerate(values):
        if not i % 5000:
            print i
            
        # Alternative way to handle messages, in the flow directly :
#        if isinstance(value, Packet):
#            print value


from setuptools import setup, find_packages
import os

version = '0.1'

setup(name='pyf.station',
      version=version,
      description="PyF.Station is a protocol with client and server implementation to transfer python generators accross tcp networks.",
      long_description=open("README.txt").read(),
      # Get more strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Programming Language :: Python",
        ],
      keywords='',
      author='',
      author_email='',
      url='http://pyfproject.org',
      license='MIT',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['pyf'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          # -*- Extra requirements: -*-
          'twisted',
          'tgscheduler',
          'pyf.transport',
          'simplejson'
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
